
function gradient(prob::Problem{T};
                 ) where {T}
    G = zeros(T, length(prob.state_vector))
    for (key, wrapper) in prob.equilibrium_wrappers
        @info "Computing the gradient at iteration $(prob.iteration)
        for $(equilibrium_type_string(wrapper)) state variables
        with method $(derivative_type_string(wrapper))"
        var_indices = findall(v -> v.equilibrium_key == key, prob.state_vector)
        eq_targets = filter(p -> prob.derived_geometries[p.second.geometry_key].parent_key == key, prob.targets)
        objective_gradient, target_gradients = gradient(deepcopy(wrapper), deepcopy(prob.state_vector[var_indices]), deepcopy(eq_targets); iteration = prob.iteration)
        foreach(p -> prob.targets[p.first].gradient = p.second, target_gradients)
        foreach(p -> G[p[2]] = objective_gradient[p[1]], enumerate(var_indices))
    end
    return G
end

struct ForwardDiff{T} <: AbstractDerivative
    δ::T
end

struct CenteredDiff{T} <: AbstractDerivative
    δ::T
end

function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target};
                  kwargs...
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: Union{ForwardDiff, CenteredDiff}}
    target_gradients = Dict{Int, Vector{T}}()
    foreach(k -> target_gradients[k] = zeros(T, length(state_variables)), keys(targets))
    objective_gradient = zeros(T, length(state_variables))
    for (i, v) in enumerate(state_variables)
        objective_derivative, target_derivatives = gradient(equilibrium_wrapper, deepcopy(v), targets)
        objective_gradient[i] = objective_derivative
        foreach(k -> target_gradients[k][i] = target_derivatives[k], keys(targets))
    end
    return objective_gradient, target_gradients     
end


function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_var::OptimizationVariable,
                  targets::Dict{Int, Target};
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: ForwardDiff}
    g = equilibrium_wrapper.grad_method
    initial_value = copy(state_var.value)
    @debug "$initial_value"
    wrapper_input = deepcopy(equilibrium_wrapper.input)
    current_values = Dict{Int, T}()
    foreach(p -> current_values[p.first] =  p.second.current_value, targets)
    target_values = empty(current_values)
    target_gradients = empty(current_values)
    objective_gradient = zero(T)
    state_var.value = initial_value + g.δ
    @debug "$state_var"
    translate!(equilibrium_wrapper, [state_var])
    @debug "$(equilibrium_wrapper.input)"
    eq = compute_equilibrium(equilibrium_wrapper)
    
    if !isnothing(eq)
        update_equilibrium!(equilibrium_wrapper, eq)
        push!(target_values, compute_targets(targets, equilibrium_wrapper)...)
        for key in keys(targets)
            target_gradients[key] = (target_values[key] - current_values[key]) / g.δ
            objective_gradient += ((targets[key].target_value - target_values[key])^2/targets[key].target_weight^2 -
                                  (targets[key].target_value - current_values[key])^2/targets[key].target_weight^2) / g.δ
        end
    else
        foreach(k -> target_gradients[k] = convert(T, NaN), keys(targets))
        objective_gradient += NaN
    end
    set_input!(equilibrium_wrapper, wrapper_input)

    return objective_gradient, target_gradients 

end


function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_var::OptimizationVariable,
                  targets::Dict{Int, Target};
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: CenteredDiff}
    g = equilibrium_wrapper.grad_method
    initial_value = copy(state_var.value)
    @debug "$initial_value"
    wrapper_input = deepcopy(equilibrium_wrapper.input)
    current_values = Dict{Int, T}()
    foreach(p -> current_values[p.first] =  p.second.current_value, targets)
    forward_targets = empty(current_values)
    backward_targets = empty(current_values)
    target_gradients = empty(current_values)
    objective_gradient = zero(T)
    state_var.value = initial_value + g.δ
    @debug "$state_var"
    translate!(equilibrium_wrapper, [state_var])
    @debug "$(equilibrium_wrapper.input)"
    eq = compute_equilibrium(equilibrium_wrapper)

    if !isnothing(eq)
        update_equilibrium!(equilibrium_wrapper, eq)
        push!(forward_targets, compute_targets(targets, equilibrium_wrapper)...)

        set_input!(equilibrium_wrapper, wrapper_input)
        state_var.value = initial_value - g.δ
        translate!(equilibrium_wrapper, [state_var])
        eq = compute_equilibrium(equilibrium_wrapper)

        if !isnothing(eq)
            update_equilibrium!(equilibrium_wrapper, eq)
            push!(backward_targets, compute_targets(targets, equilibrium_wrapper)...)
            for key in keys(targets)
                forward_objective = (targets[key].target_value - forward_targets[key])^2/targets[key].target_weight^2
                backward_objective = (targets[key].target_value - backward_targets[key])^2/targets[key].target_weight^2
                target_gradients[key] = 0.5 * (forward_targets[key] - backward_targets[key]) / g.δ
                objective_gradient += 0.5 * (forward_objective - backward_objective) / g.δ
            end
        else
            throw(ErrorException("Cannot compute centered difference gradient for 
                                  state variable $(name(state_var)), try adjusting
                                  the finite difference step size lower"))
        end
    else
        throw(ErrorException("Cannot compute centered difference gradient for 
                              state variable $(name(state_var)), try adjusting
                              the finite difference step size lower"))
    end
    set_input!(equilibrium_wrapper, wrapper_input)

    return objective_gradient, target_gradients
end

"""
Implements simultaneous perturbation stochastic approximation method
of J.C. Spall, **IEEE Trans. Automatic Control**, *37*, 3, (1992)

θₖ₊₁ = θₖ - ϵ/(β + k)^α gₖ(θ);
gₖ(θ) = f(θₖ .+ cₖ Δₖ) .- f(θₖ .- cₖ Δₖ) ./ (2 * cₖ Δ_k);
cₖ = δ/k^γ 
"""
struct SPSA{T} <: AbstractDerivative
    α::T
    β::T
    γ::T
    δ::T
    ϵ::T
    n::Int
end

function SPSA(; α = 1.0, β = 1.0, γ = 1.0, δ = 1e-4, ϵ = 1.0, n::Int = 4)
    α_p, β_p, γ_p, δ_p, ϵ_p = promote(α, β, γ, δ, ϵ)
    return SPSA{typeof(α_p)}(α_p, β_p, γ_p, δ_p, ϵ_p, n)
end

function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target};
                  iteration::Int = 1,
                  kwargs...
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: SPSA}
    g = equilibrium_wrapper.grad_method
    original_input = deepcopy(equilibrium_wrapper.input)
    target_keys = collect(keys(targets))
    θ = map(v -> v.value, state_variables)
    δθ = similar(θ)
    target_gradients = Dict{Int, Vector{T}}()
    gₖ = zeros(T, length(θ))
    gₖ_samples = zeros(T, (length(θ), g.n))
    b_dist = Bernoulli()
    cₖ = g.δ/(iteration^g.γ)
    for n in 1:g.n
        valid_eq = false
        while !valid_eq
            forward_targets = Dict{Int, T}()
            backward_targets = Dict{Int, T}()
            Δ_targets = Dict{Int, Vector{T}}()
            Δ = rand(b_dist, length(state_variables)) |> x -> cₖ * (2x .- 1)
            δθ .= θ .+ Δ
            foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
            translate!(equilibrium_wrapper, state_variables)
            forward_eq = compute_equilibrium(equilibrium_wrapper)
            if !isnothing(forward_eq)
                update_equilibrium!(equilibrium_wrapper, forward_eq)
                push!(forward_targets, compute_targets(targets, equilibrium_wrapper)...)

                set_input!(equilibrium_wrapper, original_input)
                δθ .= θ .- Δ

                foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                translate!(equilibrium_wrapper, state_variables)
                back_eq = compute_equilibrium(equilibrium_wrapper)
                if !isnothing(back_eq)
                    update_equilibrium!(equilibrium_wrapper, back_eq)
                    push!(backward_targets, compute_targets(targets, equilibrium_wrapper)...)
                    for key in target_keys
                        forward_objective = (targets[key].target_value - forward_targets[key])^2/targets[key].target_weight^2
                        backward_objective = (targets[key].target_value - backward_targets[key])^2/targets[key].target_weight^2
                        Δ_targets[key] = 0.5 * (forward_targets[key] - backward_targets[key]) ./ Δ
                        gₖ_samples[:, n] .+= 0.5 * (forward_objective - backward_objective) ./ Δ
                    end
                    mergewith!(+, target_gradients, Δ_targets)
                    valid_eq = true
                end
                set_input!(equilibrium_wrapper, original_input)
            else
                set_input!(equilibrium_wrapper, original_input)
            end
        end
    end
    foreach(key -> target_gradients[key] .*= (1/g.n), keys(target_gradients))
    aₖ = g.ϵ/((g.β + iteration)^g.α)
    gₖ[:] .= sum(gₖ_samples, dims = 2) ./ g.n
    @info "aₖ = $aₖ"
    @info "gₖ = $gₖ"
    @info "aₖ * gₖ = $(aₖ * gₖ)"
    return aₖ * gₖ, target_gradients
end