
function gradient(prob::Problem{T},
                  comms::OptComms;
                 ) where {T}
    world_rank = comms.world.rank
    world_size = comms.world.size
    compute_rank = comms.compute.rank
    G = comms.is_world_head ? zeros(T, length(prob.state_vector)) : nothing
    finished_loop = false

println("Rank $world_rank has reached the barrier in gradient function")

    if comms.is_world_head
        for rank in comms.workers
            MPI.send(true, rank, 0, comms.world.comm)
        end
    else
        MPI.recv(0, 0, comms.world.comm)
    end

println("Rank $world_rank has passed the barrier in gradient function")

    if world_rank == 0
        for (key, wrapper) in prob.equilibrium_wrappers
            @info "Computing the gradient at iteration $(prob.iteration)
            for $(equilibrium_type_string(wrapper)) state variables
            with method $(derivative_type_string(wrapper))"
            println("Computing the gradient at iteration $(prob.iteration)
            for $(equilibrium_type_string(wrapper)) state variables
            with method $(derivative_type_string(wrapper))")

            var_indices = findall(v -> v.equilibrium_key == key, prob.state_vector)
            eq_targets = filter(p -> prob.derived_geometries[p.second.geometry_key].parent_key == key, prob.targets)

            for rank in comms.workers
println("Sending continuation false code to proc $rank")

                MPI.send(false, rank, rank, comms.world.comm)

println("Sent continuation false code to proc $rank")

                MPI.send(wrapper, rank, world_size + rank, comms.world.comm)
                MPI.send(var_indices, rank, 2 * world_size + rank, comms.world.comm)
                MPI.send(eq_targets, rank, 3 * world_size + rank, comms.world.comm)
                MPI.send(prob.iteration, rank, 4 * world_size + rank, comms.world.comm)
            end

@debug "Rank $(comms.world.rank) in iteration $(prob.iteration) calling the
$(derivative_type_string(wrapper)) on wrapper $(equilibrium_type_string(wrapper))
for state variables $var_indices"

            objective_gradient, target_gradients =
                gradient(deepcopy(wrapper), deepcopy(prob.state_vector[var_indices]),
                         deepcopy(eq_targets), comms; iteration = prob.iteration)
            for (i, v) in enumerate(var_indices)
                G[v] = objective_gradient[i] 
                for (k, t) in target_gradients
                    prob.targets[k].gradient[v] = t[i]
                end
            end
        end
        for rank in comms.workers
            MPI.send(true, rank, rank, comms.world.comm)
        end
        @debug "World controller bcast true, exiting gradient"
    else

@debug "$(comms.world.rank) has finished_loop = $finished_loop"

        while !finished_loop

@debug "$(comms.world.rank) is waiting to recv"

            finished_loop, _ = MPI.recv(0, world_rank, comms.world.comm)

@debug "$(comms.world.rank) has finished_loop $finished_loop"

            if !finished_loop

@debug "$(comms.world.rank) is prepared to receive the next wrapper"

                wrapper, _ = MPI.recv(0, world_size + world_rank, comms.world.comm)

@debug "$(comms.world.rank) has received wrapper"

                var_indices, _ = MPI.recv(0, 2 * world_size + world_rank, comms.world.comm)

@debug "$(comms.world.rank) has received vector indices"

                eq_targets, _ = MPI.recv(0, 3 * world_size + world_rank, comms.world.comm)

@debug "$(comms.world.rank) has received targets"

                iteration, _ = MPI.recv(0, 4 * world_size + world_rank, comms.world.comm)

@debug "$(comms.world.rank) has received the iteration"


@debug "$(comms.world.rank) 
 in iteration $iteration calling the $(derivative_type_string(wrapper)) method on wrapper $(equilibrium_type_string(wrapper))
for state variables $var_indices"

                gradient(deepcopy(wrapper), deepcopy(prob.state_vector[var_indices]),
                         deepcopy(eq_targets), comms; iteration = iteration)
            end
        end

@debug "Rank $(comms.world.rank) exiting gradient"

    end
    return G
end

"""
    gradient(x::AbstractVector,prob::Problem,comms::OptComms)

The main function for computing the objective function and gradient value using MPI.
The global head node acts as a controller, sending worker groups which gradient index
to evaluate and gathering the results.  Only the global head node returns value of the
objective and gradient to the optimization algorithm.
"""
function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target},
                  comms::OptComms;
                  kwargs...
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: Union{ForwardDiff, CenteredDiff}}
    world_rank = comms.world.rank
    group_rank = comms.is_compute_head || comms.is_world_head ? comms.inter_group.rank : nothing
    compute_rank = comms.compute.rank
    index = zero(Int)

    if group_rank != 0
        finished_eval = false 
        # This is a worker, doing the calculations
        if comms.is_compute_head 

            recv_index_from_head_tag = 18323 + group_rank
            notify_head_tag = 21853 + group_rank
            send_result_to_head_tag = 19293 + group_rank
            send_targets_to_head_tag = 14293 + group_rank

            while !finished_eval
                index, _ = MPI.recv(0, recv_index_from_head_tag, comms.inter_group.comm)
@debug "Rank $compute_rank/$(comms.compute.size) in group $(comms.group_id) has received $index from global head"
                MPI.bcast(index, 0, comms.compute.comm)

@debug "Rank $compute_rank/$(comms.compute.size) in group $(comms.group_id) sent $index to workers"

                if index >= 1
                    # Do the calculation
                    var = state_variables[index]
                    objective_derivative, target_derivatives = 
                        gradient(equilibrium_wrapper, deepcopy(var), targets, comms.compute.comm)
                    MPI.send(group_rank, 0, notify_head_tag, comms.inter_group.comm)
                    # Send the results back to the global head
                    MPI.send(objective_derivative, 0, send_result_to_head_tag, comms.inter_group.comm)
                    # Note this is a blocking send
                    MPI.send(target_derivatives, 0, send_targets_to_head_tag, comms.inter_group.comm)

@debug "Rank $compute_rank/$(comms.compute.size) in group $(comms.group_id) sent data to global head"

                else
                    finished_eval = true
                end
            end
        else

@debug "Rank $(compute_rank+1)/$(comms.compute.size) in group $(comms.group_id) has finished_eval = $finished_eval"
            
            while !finished_eval

@debug "Rank $(compute_rank+1)/$(comms.compute.size) waiting in group $(comms.group_id) for bcast from compute head"
                
                index = MPI.bcast(nothing, 0, comms.compute.comm)
                
@debug "Rank $(compute_rank+1)/$(comms.compute.size) in group $(comms.group_id) received $index from compute head"
                
                if index >= 1
                    # Do the calculation
                    # index gets passed here, but instead split this up so that you can Mpi_send prob.equilibrium
                    var = state_variables[index]
                    _, _ = gradient(equilibrium_wrapper, deepcopy(var), targets, comms.compute.comm)
                else
                    finished_eval = true
                end
            end
        end
        return nothing
    elseif comms.is_world_head
        n_worker_groups = comms.inter_group.size - 1

        recv_result_from_worker_tags = 19293 .+ collect(1:n_worker_groups)
        recv_targets_from_worker_tags = 14293 .+ collect(1:n_worker_groups)
        notified_head_tags = 21853 .+ collect(1:n_worker_groups)
        send_index_to_worker_tags = 18323 .+ collect(1:n_worker_groups)

        gradient_indices = world_rank == 0 ? collect(1:length(state_variables)) : nothing
        target_gradients = world_rank == 0 ? Dict{Int, Vector{T}}() : nothing
        target_gradient_buffer = world_rank == 0 ? Dict{Int, Vector{T}} : nothing
        foreach(k -> target_gradients[k] = zeros(T, length(state_variables)), keys(targets))
        index_map = Dict{Int, Int}()
        G = zeros(T, length(state_variables))


        finished_eval = false
        proc_status = fill(true, n_worker_groups)
        for group_index = 1:n_worker_groups
            grad_index = findfirst(x -> x != -1, gradient_indices)
            if !isnothing(grad_index)
                    # Send out the gradient index to evaluate

@debug "Evaluating gradient no. $(grad_index) out of $(length(state_variables))"
                    
                index = gradient_indices[grad_index]
                index_map[group_index] = gradient_indices[grad_index]
                gradient_indices[grad_index] = -1

@debug "Head is sending $(grad_index)  to group_index" group_index
                    
                MPI.send(index, group_index,
                         send_index_to_worker_tags[group_index], comms.inter_group.comm)

            else
                index = -1
                MPI.send(index, group_index,
                         send_index_to_worker_tags[group_index], comms.inter_group.comm)
                proc_status[group_index] = false
            end
        end

        finished_eval = all(i -> !i, proc_status)

        while !finished_eval
            group_index, _ = MPI.recv(MPI.MPI_ANY_SOURCE, MPI.MPI_ANY_TAG, comms.inter_group.comm)
            objective_derivative, _ = MPI.recv(group_index, recv_result_from_worker_tags[group_index], comms.inter_group.comm)
            target_gradient_buffer, _ = MPI.recv(group_index, recv_targets_from_worker_tags[group_index], comms.inter_group.comm)        

@debug "Head received data from group $group_index"

            grad_index = index_map[group_index]
            G[grad_index] = objective_derivative
            for key in keys(target_gradient_buffer)
                target_gradients[key][grad_index] = target_gradient_buffer[key]
            end

            index = findfirst(x -> x != -1, gradient_indices)

            if !isnothing(index)

@debug "Head sending index $(index) out of $(length(state_variables)) to group $group_index"

                index_map[group_index] = index
                MPI.send(index, group_index, send_index_to_worker_tags[group_index], comms.inter_group.comm)
                gradient_indices[index] = -1
            else
                index = -1

@debug "Head sending index $(index) to group $group_index"

                MPI.send(index, group_index, send_index_to_worker_tags[group_index], comms.inter_group.comm)
                proc_status[group_index] = false
            end
            if all(i -> !i, proc_status)
                finished_eval = true
            end
        end
        return G, target_gradients
    end
end


function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_var::OptimizationVariable,
                  targets::Dict{Int, Target},
                  comm::MPI.Comm;
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: ForwardDiff}
    rank = MPI.Comm_rank(comm)
    g = equilibrium_wrapper.grad_method
    initial_value = copy(state_var.value)
    @debug initial_value
    wrapper_input = deepcopy(equilibrium_wrapper.input)
    current_values = rank == 0 ? Dict{Int, T}() : nothing
    target_values = rank == 0 ? empty(current_values) : nothing
    target_gradients = rank == 0 ? empty(current_values) : nothing
    objective_gradient = rank == 0 ? zero(T) : nothing
    state_var.value = initial_value + g.δ
    translate!(equilibrium_wrapper, [state_var])
    @debug "State variable $(state_var.name) changed $(initial_value) => $(equilibrium_wrapper.input[state_var.name]), $(state_var.value)"
    @debug "Rank $(rank+1)/$(MPI.Comm_size(comm)) is calling compute_equilibrium"
    eq = compute_equilibrium(equilibrium_wrapper, comm)
    
    if !isnothing(eq)
        update_equilibrium!(equilibrium_wrapper, eq)
        if rank == 0
            foreach(p -> current_values[p.first] =  p.second.current_value, targets)
            push!(target_values, compute_targets(targets, equilibrium_wrapper, comm)...)
            for key in keys(targets)
                target_gradients[key] = (target_values[key] - current_values[key]) / g.δ
                objective_gradient += ((targets[key].target_value - target_values[key])^2/targets[key].target_weight^2 -
                                  (targets[key].target_value - current_values[key])^2/targets[key].target_weight^2) / g.δ
            end
        else
            compute_targets(targets, equilibrium_wrapper, comm)
        end
    else
        if rank == 0
            foreach(k -> target_gradients[k] = convert(T, NaN), keys(targets))
            objective_gradient += NaN
        end
    end
    set_input!(equilibrium_wrapper, wrapper_input)

    return objective_gradient, target_gradients 

end


function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_var::OptimizationVariable,
                  targets::Dict{Int, Target},
                  comm::MPI.Comm;
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: CenteredDiff}
    rank = MPI.Comm_rank(comm)
    g = equilibrium_wrapper.grad_method
    initial_value = copy(state_var.value)
    wrapper_input = deepcopy(equilibrium_wrapper.input)
    current_values = rank == 0 ? Dict{Int, T}() : nothing
    forward_targets = rank == 0 ? empty(current_values) : nothing
    backward_targets = rank == 0 ? empty(current_values) : nothing
    target_gradients = rank == 0 ? empty(current_values) : nothing
    objective_gradient = rank == 0 ? zero(T) : nothing
    state_var.value = initial_value + g.δ
    translate!(equilibrium_wrapper, [state_var])
    @debug "Rank $(rank+1)/$(MPI.Comm_size(comm)) is calling compute_equilibrium with $(state_var.name = state_var.value)"
    eq = compute_equilibrium(equilibrium_wrapper, comm)

    if !isnothing(eq)
        update_equilibrium!(equilibrium_wrapper, eq)
        if rank == 0
            foreach(p -> current_values[p.first] =  p.second.current_value, targets)
            push!(forward_targets, compute_targets(targets, equilibrium_wrapper, comm)...)

            set_input!(equilibrium_wrapper, wrapper_input)
            state_var.value = initial_value - g.δ
            translate!(equilibrium_wrapper, [state_var])
            eq = compute_equilibrium(equilibrium_wrapper)

            if !isnothing(eq)
                update_equilibrium(equilibrium_wrapper, eq)
                push!(backward_targets, compute_targets(targets, equilibrium_wrapper, comm)...)
                for key in keys(targets)
                    target_gradients[key] = (forward_targets[key] - backward_targets[key]) / g.δ
                    forward_objective = (targets[key].target_value - forward_targets[key])^2/targets[key].target_weight^2
                    backward_objective = (targets[key].target_value - backward_targets[key])^2/targets[key].target_weight^2
                    objective_gradient += 0.5 * (forward_objective - backward_objective) / g.δ
                end
            else
                throw(ErrorException("Cannot compute centered difference gradient for 
                state variable $(name(state_var)), try adjusting
                the finite difference step size lower"))
            end
        else
            compute_targets(targets, equilibrium_wrapper, comm)
            set_input!(equilibrium_wrapper, wrapper_input)
            state_var.value = initial_value - g.δ
            translate!(equilibrium_wrapper, [state_var])
            eq = compute_equilibrium(equilibrium_wrapper)

            if !isnothing(eq)
                update_equilibrium(equilibrium_wrapper, eq)
                compute_targets(targets, equilibrium_wrapper, comm)
            else
                throw(ErrorException("Cannot compute centered difference gradient for 
                state variable $(name(state_var)), try adjusting
                the finite difference step size lower"))
            end
        end
    else
        throw(ErrorException("Cannot compute centered difference gradient for 
        state variable $(name(state_var)), try adjusting
        the finite difference step size lower"))
    end
    set_input!(equilibrium_wrapper, wrapper_input)

    return objective_gradient, target_gradients 

end

#=
function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target},
                  comms::OptComms;
                  iteration::Int = 1,
                  kwargs...
                ) where {T, E <: AbstractMagneticEquilibrium, D <: SPSA}
    world_rank = comms.world.rank
    compute_rank = comms.compute.rank
    group_rank = is_compute_head || is_world_head ? comms.inter_group.rank : nothing
    group_size = MPI.Comm_size(comms.inter_group.comm)

    g = equilibrium_wrapper.grad_method
    target_gradients = world_rank == 0 ? Dict{Int, Vector{T}}() : nothing
    gₖ = world_rank == 0 ? zeros(T, length(state_variables)) : nothing
    gₖ_samples = world_rank == 0 ? zeros(T, (length(state_variables), g.n)) : nothing

    if world_rank == 0
        group_status = fill(true, group_size-1)
        sample_status = fill(false, g.n) 
        for i in eachindex(group_status)
            sample_id = findfirst(!j, sample_status)
            if !isnothing(sample_id)
                MPI.send(true, i, i, comms.inter_group.comm)
            else
                MPI.send(false, i, i, comms.inter_group.comm)
                group_status[i] = false
            end
        end
        finished = all(s, sample_status) && group_status(!s, group_status)
        while !finished
            index, _ = MPI.recv(MPI.MPI_ANY_SOURCE, MPI.MPI_ANY_TAGS, comms.inter_group.comm)
            objective_derivatives, _ = MPI.recv(index, group_size + index, comms.inter_group.comm) 
            target_derivatives, _ = MPI.recv(index, 2 * group_size + index, comms.inter_group.comm) 
            sample_id = findfirst(!s, sample_status)
            gₖ_samples[:, sample_id] .= objective_derivatives
            mergewith!(+, target_gradients, target_derivatives)
            sample_status[sample_id] = true

             
            sample_id = findfirst(!s, sample_status)
            if !isnothing(sample_id)
                MPI.send(true, index, index, comms.inter_group.comm)
            else
                MPI.send(false, index, index, comms.inter_group.comm)
                group_status[index] = false
            end
            
            finished = all(s, sample_status) && group_status(!s, group_status)
        end
    else
        if compute_rank == 0
            finished = false
            while !finished
                compute_flag, _ = MPI.recv(0, group_rank, comms.inter_group.comm)
                MPI.bcast(compute_flag, 0, comms.compute.comm)
                if compute_flag
                    objective_derivative, target_derivatives = 
                        gradient(equilibrium_wrapper, state_variables, targets, 
                                 comms.compute.comm, iteration = iteration)
                    MPI.send(group_rank, 0, group_rank, comms.inter_group.comm)
                    MPI.send(objective_derivative, 0, group_size + group_rank, comms.inter_group.comm)
                    MPI.send(target_derivatives, 0, 2 * group_size + group_rank, comms.inter_group.comm)
                else
                    finished = true
                end
            end
        else
            finished = false
            while !finished
                compute_flag = MPI.bcast(nothing, 0, comms.compute.comm)
                if compute_flag
                    gradient(equilibrium_wrapper, state_variables, targets,
                             comms.compute.comm, iteration = iteration)
                else
                    finished = true
                end
            end
        end
    end
    if world_rank == 0
        foreach(key -> target_gradients[key] .*= (1/g.n), keys(target_gradients))
        @debug "Final target_gradients: $target_gradients"
        aₖ = g.ϵ/((g.β + iteration)^g.α)
        gₖ[:] .= sum(gₖ_samples, dims = 2) ./ g.n
        @info "aₖ = $aₖ"
        @info "gₖ = $gₖ"
        @info "aₖ * gₖ = $(aₖ * gₖ)"
        return aₖ * gₖ, target_gradients
    else
        return nothing
    end
end    

function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target},
                  comm::MPI.Comm;
                  iteration::Int = 1,
                  kwargs...
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: SPSA}
    rank = MPI.Comm_rank(comm)
    g = equilibrium_wrapper.grad_method
    original_input = deepcopy(equilibrium_wrapper.input)
    target_keys = rank == 0 ? collect(keys(targets)) : nothing
    θ = map(v -> v.value, state_variables)
    δθ = similar(θ)
    target_gradients = rank == 0 ? Dict{Int, Vector{T}}() : nothing
    gₖ = rank == 0 ? zeros(T, length(θ)) : nothing
    b_dist = rank == 0 ? Bernoulli() : nothing
    cₖ = rank == 0 ? g.δ/(iteration^g.γ) : nothing
    valid_eq = false
    while !valid_eq
        if rank == 0
            forward_targets = Dict{Int, T}()
            backward_targets = Dict{Int, T}()
            Δ = rand(b_dist, length(state_variables)) |> x -> cₖ * (2x .- 1)
            δθ .= θ .+ Δ
            MPI.bcast(δθ, 0, comm)
            @debug "Rank $rank has forward perturbations $δθ"
            #foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
            for i in 1:length(δθ)
                state_variables[i].value = δθ[i]
                @debug "State variable $(name(state_variables[i])) on rank $rank has value $(state_variables[i].value)"
            end
            translate!(equilibrium_wrapper, state_variables)
            forward_eq = compute_equilibrium(equilibrium_wrapper, comm)
            if !isnothing(forward_eq)
                update_equilibrium!(equilibrium_wrapper, forward_eq)
                push!(forward_targets, compute_targets(targets, equilibrium_wrapper, comm)...)

                set_input!(equilibrium_wrapper, original_input)
                δθ .= θ .- Δ
                MPI.bcast(δθ, 0, comm)
                @debug "Rank $rank has backward perturbations $δθ"

                #foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                for i in 1:length(δθ)
                    state_variables[i].value = δθ[i]
                    @debug "State variable $(name(state_variables[i])) on rank $rank has value $(state_variables[i].value)"
                end
                translate!(equilibrium_wrapper, state_variables)
                back_eq = compute_equilibrium(equilibrium_wrapper, comm)
                if !isnothing(back_eq)
                    update_equilibrium!(equilibrium_wrapper, back_eq)
                    push!(backward_targets, compute_targets(targets, equilibrium_wrapper, comm)...)
                    for key in target_keys
                        forward_objective = (targets[key].target_value - forward_targets[key])^2/targets[key].target_weight^2
                        backward_objective = (targets[key].target_value - backward_targets[key])^2/targets[key].target_weight^2
                        target_gradients[key] = 0.5 * (forward_targets[key] - backward_targets[key]) ./ Δ
                        gₖ[:] .+= 0.5 * (forward_objective - backward_objective) ./ Δ
                    end
                    @debug "target_gradients: $target_gradients"
                    @debug "gₖ = $gₖ"
                    valid_eq = true
                end
                set_input!(equilibrium_wrapper, original_input)
            else
                set_input!(equilibrium_wrapper, original_input)
            end
            MPI.bcast(valid_eq, 0, comm)
        else
            δθ = MPI.bcast(nothing, 0, comm)
            @debug "Rank $rank has forward perturbations $δθ"
            #oreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
            for i in 1:length(δθ)
                state_variables[i].value = δθ[i]
                @debug "State variable $(name(state_variables[i])) on rank $rank has value $(state_variables[i].value)"
            end
            translate!(equilibrium_wrapper, state_variables)
            forward_eq = compute_equilibrium(equilibrium_wrapper, comm)
            if !isnothing(forward_eq)
                update_equilibrium!(equilibrium_wrapper, forward_eq)
                compute_targets(targets, equilibrium_wrapper, comm)

                set_input!(equilibrium_wrapper, original_input)
                δθ = MPI.bcast(nothing, 0, comm)
                @debug "Rank $rank has backward perturbations $δθ"

                #foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                for i in 1:length(δθ)
                    state_variables[i].value = δθ[i]
                    @debug "State variable $(name(state_variables[i])) on rank $rank has value $(state_variables[i].value)"
                end

                translate!(equilibrium_wrapper, state_variables)
                back_eq = compute_equilibrium(equilibrium_wrapper, comm)
                if !isnothing(back_eq)
                    update_equilibrium!(equilibrium_wrapper, back_eq)
                    compute_targets(targets, equilibrium_wrapper, comm)
                end
                set_input!(equilibrium_wrapper, original_input)
            else
                set_input!(equilibrium_wrapper, original_input)
            end
            valid_eq = MPI.bcast(nothing, 0, comm)
        end
    end
    if rank == 0
        return gₖ, target_gradients
    else
        return nothing
    end
end
=#
