using .MPI


"""
    OptComms

Composite type holding the different communicators for MPI-based optimization.

# Fields:
- world_comm::MPI.Comm  # The top-level communicator between all processors. Rank 0 on this communicator is the global head.
- group_comm::MPI.Comm  # Communicator linking all the group heads.  Rank 0 on this communicator is the global head.
- compute_comm::MPI.Comm  # Communicator linking all the processors in a particular group.  Rank 0 on this communicator is the group head.
"""
#=
struct OptComms
    world_comm::MPI.Comm
    group_comm::MPI.Comm
    compute_comm::MPI.Comm
    ranks::NamedTuple{(:world, :group, :compute), Tuple{Int, Int, Int}}
    is_worker::Bool
    workers::Vector{Int32}
end
=#

const CommInfo = NamedTuple{(:comm, :group, :rank, :size), Tuple{MPI.Comm, MPI.Group, Int, Int}}

struct OptComms
    world::CommInfo
    inter_group::CommInfo
    compute::CommInfo
    group_id::Int
    is_world_head::Bool
    is_compute_head::Bool
    is_worker::Bool
    workers::Vector{Int32}
end

"""
    setup_mpi_comms(procs_per_group, procs_per_node, world_comm::MPI.Comm)

Split `world_comm` into groups where the communicator splitting attempts to
adhere as close to `procs_per_group` as possible, returning an `OptComms` type.

See also: [`OptComms`](@ref)
"""
function setup_mpi_comms(procs_per_group::Int,
                         procs_per_node::Int,
                         world_comm::MPI.Comm)
    world_size = MPI.Comm_size(world_comm)
    world_rank = MPI.Comm_rank(world_comm)
    world_group = MPI.Comm_group(world_comm)
    
    # Partition the workers into different groups
    group_div, group_rem = divrem(procs_per_node-1, procs_per_group)
    n_groups_per_node = group_rem == 0 ? group_div : group_div +1
    n_nodes = div(world_size, procs_per_node)
    n_worker_groups = n_nodes * n_groups_per_node

    groups = Vector{MPI.Group}(undef, n_worker_groups)
    group_ranks = Vector{Vector{Int32}}(undef, n_worker_groups)
    group_comms = Vector{MPI.Comm}(undef, n_worker_groups)

    n_procs_group = group_rem == 0 ? collect(1:n_groups_per_node) .* procs_per_group : group_rem .+ (collect(0:n_groups_per_node-1) .* procs_per_group)
    push!(n_procs_group, 0)
    n_procs_group = circshift(n_procs_group, 1)

    group_id = 0
    for r in 2:n_groups_per_node+1
        for n in 1:n_nodes
            group_ranks[n_groups_per_node * (n-1) + r-1] = Int32.((procs_per_node * (n - 1)) .+ findall(i -> n_procs_group[r-1] < i <= n_procs_group[r], collect(1:world_size-1)))
        end
    end

    for (i, r) in enumerate(group_ranks)
        group_id = iszero(group_id) ? first(r) <= world_rank <= last(r) ? i : 0 : group_id 
    end
    is_worker = !iszero(group_id)
   
    for i in 1:n_worker_groups
        groups[i] = MPI.Group_incl(world_group, group_ranks[i])
        group_comms[i] = MPI.Comm_create_group(world_comm, groups[i], i)
    end

    workers = Vector{Int32}(undef, 0)
    for i in 1:n_worker_groups
        workers = vcat(workers, group_ranks[i])
    end
   
    head_ranks = Int32.(circshift(push!(first.(group_ranks), 0), 1))
    head_group = MPI.Group_incl(world_group, head_ranks)
    group_comm = MPI.Comm_create_group(world_comm, head_group, 0)
    is_compute_head = world_rank in head_ranks
    group_rank = is_compute_head ? MPI.Comm_rank(group_comm) : -1
    group_size = is_compute_head ? MPI.Comm_size(group_comm) : -1

    if group_id != 0
        compute_comm = group_comms[group_id]
        compute_group = groups[group_id]
        compute_rank = MPI.Comm_rank(compute_comm)
        compute_size = MPI.Comm_size(compute_comm)
    elseif world_rank == 0
        compute_comm = MPI.COMM_NULL
        compute_group = MPI.GROUP_NULL
        compute_rank = -1
        compute_size = -1
    else
        compute_comm = MPI.COMM_NULL
        compute_group = MPI.GROUP_NULL
        compute_rank = -1
        compute_size = -1
        group_id = -1
    end

    
    #=
    ranks = NamedTuple{(:world, :group, :compute), Tuple{Int, Int, Int}}((world_rank,group_id,compute_rank))
    return OptComms(world_comm, group_comm, compute_comm, ranks, is_worker, workers)
    =#
    @debug OptComms((comm = world_comm, group = world_group, rank = world_rank, size = world_size),
                    (comm = group_comm, group = head_group, rank = group_rank, size = group_size),
                    (comm = compute_comm, group = compute_group, rank = compute_rank, size = compute_size),
                    group_id, is_compute_head && !is_worker, is_compute_head, is_worker, workers)
    return OptComms((comm = world_comm, group = world_group, rank = world_rank, size = world_size),
                    (comm = group_comm, group = head_group, rank = group_rank, size = group_size),
                    (comm = compute_comm, group = compute_group, rank = compute_rank, size = compute_size),
                    group_id, is_compute_head && !is_worker, is_compute_head, is_worker, workers)
end


"""
    @add_mpi_target!(prob, method, target, weight, label)

Construct a target from a method signature provided in `method.`  The
method signature *must* have keyword arugment `geometry = geometry_key`,
where `geometry_key` is the identifier returned by `@add_derived_geometry`.

Example:
```julia-repl
julia> using MPI, VMEC, StellaratorOptimization

julia> prob = StellOpt.Problem{Float64}();

julia> vmec_wrapper = EquilibriumWrapper(input_dictionary);

julia> add_equilibrium_wrapper!(prob, vmec_wrapper);

julia> vmec_surface = @add_derived_geometry!(prob, VmecSurface(0.5, equilibrium = vmec_wrapper));

julia> @add_mpi_target!(prob, VMEC.quasisymmetry_deviation(1, 4, 16, 16, geometry = vmec_surface), 0.0, 1.0, "qs(0.5)");
```
"""
macro add_mpi_target!(prob, method, target, weight, label)
  mod = @__MODULE__

  #method_call = Expr(method.head, method.args...)
  #sym, index = find_last_module(method)
  #method_call.args[1] = insert_main(method, sym, index)
  method_expr = Expr(:quote, method)
  return esc(quote
        # Need to evaluate an argument in the calling scope
        # so the values can be passed to the calling function
        # The first argument will always be the function to be called
        method_call = Expr($method_expr.head, $method_expr.args...)
        calling_mod, index = $mod._find_calling_module(method_call)
        escaped_mod = Symbol(@__MODULE__)
        method_call = $mod._insert_escaped_module(method_call, escaped_mod, calling_mod, index)
        for i in 2:length($method_expr.args)
            if Meta.isexpr($method_expr.args[i], :kw)
                method_call.args[i].args[2] = eval($method_expr.args[i].args[2])
            else
                method_call.args[i] = eval($method_expr.args[i])
            end
        end
        geometry_slot = findfirst(x -> x isa UInt, method_call.args)
        geometry_key = !isnothing(geometry_slot) ? method_call.args[geometry_slot] : error("Target must have a UInt identifier for the geometry")
        comm_slot = findfirst(x -> x isa $mod.OptComms, method_call.args)
        if !isnothing(comm_slot)
            method_call.args[comm_slot] = :mpi_comm
        end
        $mod.add_target!($prob, $mod.Target($target, $weight, method_call, geometry_key, $label))
  end)
  
end
