
"""
    optimize_mpi(prob::Problem,comms::OptComms)

Run an optimization problem defined by `prob` using MPI with communicators
defined in `comms`.
"""
function optimize_mpi(prob::Problem{T},
                      comms::OptComms;
                     ) where {T}
    if comms.is_world_head
        # Define the callback function that requires knowledge of the problem
        function cb_function(x)
            update_history!(x, prob)
        end

        # Extract initial state vector of BitType data
        initial_state_vector = state_vector_values(prob)

        # Set the intial gradient of each target to zero
        for i in 1:length(prob.targets)
            prob.targets[i].gradient = zeros(T, length(prob.state_vector))
        end

        # Augment the arguments of the optimizer options with the callback
        option_args = Dict{Symbol, Any}(:callback => cb_function)
        for f in fieldnames(Optim.Options)
            val = getfield(prob.optimizer.options, f)
            if !isnothing(val)
                option_args[f] = val
            end
        end
        if !option_args[:extended_trace]
            option_args[:extended_trace] = true
        end
        if !option_args[:store_trace]
            option_args[:store_trace] = true
        end
        new_options = (; option_args...)

        write_history_attributes(prob)  
        log_io = open(joinpath(prob.output_directory, "stellopt_$(prob.output_tag)_log.jl"), "w+")
        logger = SimpleLogger(log_io, Logging.Info)
        residual = with_logger(logger) do
            Optim.optimize(Optim.only_fg!((F, G, θ) -> objective_gradient!(F, G, θ, prob, comms; logger_io = log_io)),
                           initial_state_vector, prob.optimizer.method, Optim.Options(; new_options...))
        end
        flush(log_io)
        close(log_io)
        #write_history_file(prob)
        for i in comms.workers
            MPI.send(true, i, i, comms.world.comm)
        end
        return residual
    else
        if comms.is_worker
            worker_loop(prob, comms)
        end
        return nothing
    end
end

function objective_gradient!(F,
                             G,
                             x::AbstractVector,
                             prob::Problem,
                             comms::OptComms;
                             logger_io::Union{Nothing, IOStream} = nothing,
                            )
    # If the loop is in this function there is still work left to do; bcast false
println(comms.workers)
    for i in comms.workers
println(i)
println("Head sending false to $i")
        MPI.send(false, i, i, comms.world.comm)
    end
    @debug "World controller has signaled the workers to continue"
    for i in comms.workers
        MPI.send(x, i, i, comms.world.comm)
    end
    adjust_state_vector!(prob, x)
    obj, grad = (Inf, nothing)
    if isnothing(F)
        for i in comms.workers
            MPI.send(false, i, i, comms.world.comm)
        end
    else
        for i in comms.workers
            MPI.send(true, i, i, comms.world.comm)
        end
        @debug "Calling compute_objective from the controller"
        obj = compute_objective(prob, comms)
        for i in comms.workers
            MPI.send(obj, i, i, comms.world.comm)
        end
println("Controller has bcast objective $obj to workers")
        adjust_objective!(prob, obj)
    end
    if isnothing(G)
        for i in comms.workers
            MPI.send(false, i, i, comms.world.comm)
        end
    else
        if !isnan(obj)
            for i in comms.workers
                MPI.send(true, i, i, comms.world.comm)
            end
            grad = gradient(prob, comms)
            copyto!(G, grad)
            adjust_gradient!(prob, grad)
        else
            for i in comms.workers
                MPI.send(false, i, i, comms.world.comm)
            end
            copyto!(G, fill(NaN, length(x)))
        end
    end
    !isnothing(logger_io) ? flush(logger_io) : nothing
    if !isnothing(F)
        return obj
    end
    nothing
end

"""
    worker_loop(prob::Problem,comms::OptComms)

Listens for optimization state vector and calculation termination from the
global head node.  Exits when a finished value is broadcast from the global
head node.
"""
function worker_loop(prob::Problem,
                     comms::OptComms;
                    )
    finished = false

    while !finished
        # Check whether head is broadcasting if optimization is finished
        finished, _ = MPI.recv(0, comms.world.rank, comms.world.comm)
        @debug "Global rank $(comms.world.rank) has recieved $finished from bcast"
        if !finished
            # Get updated state vector values from head node
            state_vector, _ = MPI.recv(0, comms.world.rank, comms.world.comm)
            @debug "Global rank $(comms.world.rank) has received new state vector"
            adjust_state_vector!(prob, state_vector)
            # Check whether objective function calculation is required
            objective_needed, _ = MPI.recv(0, comms.world.rank, comms.world.comm)
            if objective_needed

@debug "$(comms.world.rank) is calling compute_objective"

                compute_objective(prob, comms)

@debug "$(comms.world.rank) has exited compute_objective"

                obj, _ = MPI.recv(0, comms.world.rank, comms.world.comm)

@debug "$(comms.world.rank) has received the objective $obj"

                adjust_objective!(prob, obj)

@debug "$(comms.world.rank) has adjusted the objective"

            end
            # Check whether gradient calculation is required
            gradient_needed, _ = MPI.recv(0, comms.world.rank, comms.world.comm)

@debug "$(comms.world.rank) has received gradient_needed = $gradient_needed"

            if gradient_needed

@debug "$(comms.world.rank) is calling gradient"

                gradient(prob, comms)
            end
        end
    end
end

#TODO: Clean this up
function compute_objective(prob::Problem{T},
                           comms::OptComms;
                          ) where {T}
    @debug "Global rank $(comms.world.rank) is in compute_objective"
    world_size = comms.world.size

    # Compute the initial equilibrium and objective functions, broadcast
    # the equilibrium data to all processors. At this point all ranks
    # have the equilibrium input data from the previous iteration, thus
    # can immediately call the evaluate function
    if comms.group_id == 0
        F = zero(T)
        success, _ = MPI.recv(1, comms.world.rank, comms.world.comm)
        if success
            for i in keys(prob.equilibrium_wrappers)
                eq_data, _ = MPI.recv(1, comms.world.rank, comms.world.comm)
                update_equilibrium!(prob.equilibrium_wrappers[i], eq_data)
                update_input!(prob.equilibrium_wrappers[i])
            end
            target_values, _ = MPI.recv(1, 1, comms.world.comm)
            foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
            F, _ = MPI.recv(1, 2, comms.world.comm)
        else
            F = convert(T, NaN)
        end
        return F
    elseif comms.group_id == 1
        send_ids = comms.is_compute_head ? push!(filter(i -> i != comms.world.rank, comms.workers), 0) : nothing
        valid_eqs = fill(true, length(prob.equilibrium_wrappers))
        j = 0
        # Do the calculation on the first worker group
        for i in keys(prob.equilibrium_wrappers)
            j += 1
            translate!(prob.equilibrium_wrappers[i], prob.state_vector)

@debug "$(comms.world.rank) is calling compute_equilibrium"

            eq = compute_equilibrium(prob.equilibrium_wrappers[i], comms.compute.comm)
            if !isnothing(eq) && typeof(eq) <: equilibrium_type(prob.equilibrium_wrappers[i])
                update_equilibrium!(prob.equilibrium_wrappers[i], eq)
                update_input!(prob.equilibrium_wrappers[i])
            else
                valid_eqs[j] = false
            end
        end

        if all(valid_eqs)
            if comms.is_compute_head
                for i in send_ids
                    MPI.send(true, i, i, comms.world.comm)
                end
            else
                MPI.recv(1, comms.world.rank, comms.world.comm)
            end

            for j in keys(prob.equilibrium_wrappers)
                if comms.is_compute_head
                    for i in send_ids
                        MPI.send(prob.equilibrium_wrappers[j].eq, i, i, comms.world.comm)
                    end
                else
                    MPI.recv(1, comms.world.rank, comms.world.comm)
                end
            end
            target_values = compute_all_targets(prob, comms.compute.comm, group_id = comms.group_id)

            target_values = MPI.bcast(target_values, 0, comms.compute.comm)
            if comms.is_compute_head
                for i in 1:comms.compute.size-1
                    MPI.send(target_values, i, i, comms.compute.comm)
                end
            else
                target_values, _ = MPI.recv(0, comms.compute.rank, comms.compute.comm)
            end

            foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
            if comms.is_compute_head
                F = least_squares_objective(prob)
                MPI.send(target_values, 0, 1, comms.world.comm) 
                MPI.send(F, 0, 2, comms.world.comm)
            end
        else
            if comms.is_compute_head
                for i in send_ids
                    MPI.send(false, i, i, comms.world.comm)
                end
            else
                MPI.recv(1, comms.world.rank, comms.world.comm)
            end

        end

@debug "$(comms.world.rank) is exiting compute_objective"

        return nothing
    else
        success, _ = MPI.recv(1, comms.world.rank, comms.world.comm)

@debug "$(comms.world.rank) with comm $(comms.compute.comm) has received success code $success"

        if success 
            for i in keys(prob.equilibrium_wrappers)
                eq, _ = MPI.recv(1, comms.world.rank, comms.world.comm)
                update_equilibrium!(prob.equilibrium_wrappers[i], eq)
                update_input!(prob.equilibrium_wrappers[i])
            end

@debug "$(comms.world.rank) is calling compute_all_targets with comm $(comms.compute.comm)"

            target_values = compute_all_targets(prob, comms.compute.comm; group_id = comms.group_id)

@debug "$(comms.world.rank) has exited compute_all_targets"
               
            target_values = MPI.bcast(target_values, 0, comms.compute.comm)

@debug "$(comms.world.rank) has received the target values"

            foreach(p -> prob.targets[p.first].current_value = p.second, target_values)

@debug "$(comms.world.rank) has assigned new target values"

        end

@debug "$(comms.world.rank) is exiting compute_objective"

        return nothing
    end
end

