
function optimize_threads(prob::Problem{T}) where {T}
    # Define the callback function that requires knowledge of the problem
    function cb_function(x)
        @debug "Executing the callback at iteration $(prob.iteration)"
        update_history!(x, prob)
    end

    # Extract initial state vector of BitType data
    initial_state_vector = state_vector_values(prob)

    # Set the intial gradient of each target to zero
    for i in 1:length(prob.targets)
        prob.targets[i].gradient = zeros(T, length(prob.state_vector))
    end

    # Augment the arguments of the optimizer options with the callback
    option_args = Dict{Symbol, Any}(:callback => cb_function)
    for f in propertynames(prob.optimizer.options)
        val = getfield(prob.optimizer.options, f)
        if !isnothing(val)
            option_args[f] = val
        end
    end
    if !haskey(option_args, :extended_trace)
        option_args[:extended_trace] = true
    end
    if !haskey(option_args, :store_trace)
        option_args[:store_trace] = true
    end
    new_options = (; option_args...)

    write_history_attributes(prob)
    log_io = open(joinpath(prob.output_directory, "stellopt_$(prob.output_tag)_log.jl"), "w+")
    logger = SimpleLogger(log_io, Logging.Info)
    residual = with_logger(logger) do
         Optim.optimize(Optim.only_fg!((F, G, θ) -> objective_gradient!(F, G, θ, prob; logger_io = log_io)),
                        initial_state_vector, prob.optimizer.method, Optim.Options(; new_options...))
    end
    flush(log_io)
    close(log_io)
    #write_history_file(prob)
    return residual
end

function objective_gradient!(F,
                             G,
                             θ::AbstractVector,
                             prob::Problem;
                             logger_io::Union{Nothing, IOStream} = nothing,
                            )
    @debug "Calling objective_gradient at iteration $(prob.iteration)"
    adjust_state_vector!(prob, θ)
    obj, grad = (Inf, nothing)
    @debug "isnothing(F): $(isnothing(F)), isnothing(G): $(isnothing(G))"
    if !isnothing(F)
        obj = compute_objective(prob, !isnothing(G))
        adjust_objective!(prob, obj)
    end
    if !isnothing(G)
        if !isnan(obj)
            grad = gradient(prob)
            copyto!(G, grad)
            adjust_gradient!(prob, grad)
        else
            copyto!(G, fill(NaN, length(θ)))
        end
    end
    !isnothing(logger_io) ? flush(logger_io) : nothing
    if !isnothing(F)
        return obj
    end
    nothing
end

function compute_objective(prob::Problem,
                           new_iter::Bool = true;
                          )
    valid_eqs = fill(true, length(prob.equilibrium_wrappers))
    j = 0
    for i in eachindex(prob.equilibrium_wrappers)
        j += 1
        translate!(prob.equilibrium_wrappers[i], prob.state_vector)
        eq = compute_equilibrium(prob.equilibrium_wrappers[i])
        if !isnothing(eq)
            update_equilibrium!(prob.equilibrium_wrappers[i], eq)
        else
            valid_eqs[j] = false
        end
    end

    if all(valid_eqs)
        compute_all_targets!(prob)
        F = least_squares_objective(prob)
        return F
    else
        return NaN
    end
end

function least_squares_objective(prob::Problem{T};
                                 gradient_index::Union{Nothing, Int} = nothing,
                                 finite_diff_step::Union{Nothing, Int} = nothing
                                ) where {T}
    obj = zero(T)
    for (index, target) in prob.targets
        #length(target.label) < 25 ? target_label = target.label * repeat(" ", 25-length(target.label)) : target_label = target.label
        #target_string = @sprintf "%.3E" target.target_value
        #value_string = @sprintf "%.5E" target.current_value
        #sigma_string = @sprintf "%.3E" target.target_weight
        #println(target_label,": target ",target_string," value ",value_string," sigma ",sigma_string)
        if isa(target.target_value, Tuple) || isa(target.target_value, AbstractVector)
            for i = 1:length(target.target_value)
                obj += (target.target_value[i] - target.current_value[i])^2 / (target.target_weight[i]^2)
            end
        else
            obj += (target.target_value - target.current_value)^2 / (target.target_weight^2)
        end
    end
    objective_string = @sprintf "%.5E" obj
    #println("Objective value: ",objective_string)
    return obj
end
