
function compute_all_targets!(prob::Problem,
                              comm::MPI.Comm;
                              kwargs...
                             )
    rank = MPI.Comm_rank(comm)
    if rank == 0
        target_values = compute_all_targets(prob, comm; kwargs...)
        foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
    else
        compute_all_targets(prob, comm; kwargs...)
    end
end

function compute_all_targets(prob::Problem,
                             comm::MPI.Comm;
                             kwargs...
                            )
    return compute_targets(prob.targets, prob.equilibrium_wrappers, comm; kwargs...)
end

function compute_targets(targets::Dict{Int, Target},
                         equilibrium_wrappers::Union{EquilibriumWrapper, Dict{UInt, EquilibriumWrapper}},
                         comm::MPI.Comm;
                         group_id = 0
                        )
    rank = MPI.Comm_rank(comm)
    comm_size = MPI.Comm_size(comm)
    if rank == 0
        T = typeof(first(targets).second.target_value)
        target_values =Dict{Int, T}()
        wrappers = equilibrium_wrappers isa Dict ? equilibrium_wrappers : Dict(one(UInt) => equilibrium_wrappers)
        for eq_wrapper in values(wrappers)
            for key in keys(eq_wrapper.derived_geometries)
                geometry = extract_derived_geometry(eq_wrapper, key)
                geometry_targets = filter(t -> t.second.geometry_key == key, targets)
                for i in 1:comm_size-1
@debug "Rank $rank in group $group_id is sending false"
                    MPI.send(false, i, i, comm)
@debug "Rank $rank in group $group_id is sending the key"
                    MPI.send(key, i, comm_size + i, comm)
@debug "Rank $rank in group $group_id is sending the geometry"
                    MPI.send(geometry, i, 2 * comm_size + i, comm)
@debug "Rank $rank in group $group_id is sending the targets"
                    MPI.send(geometry_targets, i, 3 * comm_size + i, comm)
                end
                #@debug "Rank $rank on comm $comm is calling targets with geometry $key"
                results = compute_targets(deepcopy(geometry_targets), geometry, comm; group_id = group_id)
                foreach(p -> target_values[p.first] = p.second, results)
            end
        end
        for i in 1:comm_size-1
            MPI.send(true, i, i, comm)
        end
        return target_values
    else
        finished = false
        while !finished
            finished, _ = MPI.recv(0, rank, comm)
@debug "Rank $rank in group $group_id has received $finished"
            if !finished
                key, _ = MPI.recv(0, comm_size + rank, comm)
@debug "Rank $rank in group $group_id has received the key"
                geometry, _ = MPI.recv(0, 2 * comm_size + rank, comm)
@debug "Rank $rank in group $group_id has received the geometry"
                geometry_targets, _ = MPI.recv(0, 3 * comm_size + rank, comm)
@debug "Rank $rank in group $group_id has received the targets"
                #@debug "Rank $rank on comm $comm is calling targets with geometry $key"
                compute_targets(geometry_targets, geometry, comm; group_id =group_id)
            end
        end
    end
end
        

function compute_targets(targets::Dict{Int, Target},
                         geometry::E,
                         comm::MPI.Comm;
                         group_id = 0,
                        ) where {E <: AbstractGeometry}
    rank = MPI.Comm_rank(comm)
    comm_size = MPI.Comm_size(comm)
    if rank == 0
        T = typeof(first(targets).second.target_value)
        target_values = Dict{Int, T}()
        for (i, target) in targets
            for r in 1:comm_size-1
@debug "Rank $rank in group $group_id is sending false"
                MPI.send(false, r, r, comm)
@debug "Rank $rank in group $group_id is sending the target"
                MPI.send(target, r, comm_size + r, comm)
            end
            #@debug "Rank $rank on comm $comm is calling $(target.label)"
            target_values[i] = compute_target(target, geometry, comm; group_id = group_id)
        end
        for r in 1:comm_size-1
            MPI.send(true, r, r, comm)
        end
        return target_values
    else
        finished = false
        while !finished
            finished, _ = MPI.recv(0, rank, comm)
@debug "Rank $rank in group $group_id has received the $finished"
            if !finished
                target, _ = MPI.recv(0, comm_size + rank, comm)
@debug "Rank $rank in group $group_id has received the target"
                #@debug "Rank $rank on comm $comm is calling $(target.label)"
                compute_target(target, geometry, comm; group_id = group_id)
            end
        end
    end
end

function compute_target(target::Target,
                        geometry::E,
                        comm::MPI.Comm;
                        group_id = 0
                       ) where {T, E <: AbstractGeometry}
    rank = MPI.Comm_rank(comm)
    comm_size = MPI.Comm_size(comm)
    if rank == 0
        has_comm = :mpi_comm in target.method.args
@debug "Rank $rank in group $group_id is sending $has_comm"
        for i in 1:comm_size-1
            MPI.send(has_comm, i, i, comm)
        end
        target_method = Expr(target.method.head, 
                             replace(target.method.args, 
                                     target.geometry_key => geometry,
                                     :mpi_comm => comm)...)
        push!(target_method.args, Expr(:kw, :group_id, group_id))
@debug "Rank $rank in group $group_id is sending the target"
        for i in 1:comm_size-1
            MPI.send(target_method, i, i, comm)
        end
@debug "Rank $rank in group $group_id is calling $(target_method.args[1])"
        return eval(target_method)
@debug "Rank $rank in group $group_id finished calling $(target_method.args[1])"
    else
        has_comm, _ = MPI.recv(0, rank, comm)
@debug "Rank $rank in group $group_id has received $has_comm"
        target_method, _ = MPI.recv(0, rank, comm)
@debug "Rank $rank in group $group_id has received the target"
        if has_comm
@debug "Rank $rank in group $group_id is calling $(target_method.args[1])"
            eval(target_method)
@debug "Rank $rank in group $group_id finished calling $(target_method.args[1])"
        end
    end
end
